/*
 * eth_ecat_pdo_fmmu.hpp
 *
 *  Created on: May 16, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_

#include <vector>

#include "free_rtos/semaphore/semaphore.hpp"
#include "free_rtos/mutex/mutex.hpp"

#include "free_rtos/ethernet_industry/eth_ecat_buffer.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_custom_promise.hpp"
#include "free_rtos/profiler/quick_profiler.hpp"

namespace free_rtos {

namespace ecat_pdo_fmmu {

#define PROFILE_LEVEL 0
#define PROFILE_SAMPLES 64

typedef void (*ProcessCallback)();

class EcatPdoFMMUSlave {
public:
    EcatPdoFMMUSlave(ecat_buffer::EcatBufferSlave& buffer_slave)
    : buffer_slave_(buffer_slave) { }

    template<typename DataType>
    datagram::EcatDatagram<command::LWR, DataType> make_datagram_write(DataType& data) {
        using TDatagram = datagram::EcatDatagram<command::LWR, DataType>;

        ecat_buffer::FMMUProperties& fmmu_properties = buffer_slave_.get_fmmu_properties_write();
        address::Logical logical = fmmu_properties.address;

        return TDatagram{ {{logical}}, data };
    }

    template<typename DataType>
    datagram::EcatDatagram<command::LRD, DataType> make_datagram_read(DataType& data) {
        using TDatagram = datagram::EcatDatagram<command::LRD, DataType>;

        ecat_buffer::FMMUProperties& fmmu_properties = buffer_slave_.get_fmmu_properties_read();
        address::Logical logical = fmmu_properties.address;

        return TDatagram{ {{logical}}, data };
    }

private:
    ecat_buffer::EcatBufferSlave& buffer_slave_;
};

class EthEcatPdoFMMU {
public:
    EthEcatPdoFMMU(ecat_buffer::EthEcatBuffer& ecat_buffer): ecat_buffer_{ecat_buffer}
    { }

    void init(ProcessCallback callback = nullptr);

    template<class DataT, class Allocator>
    void process(Allocator& allocator);

    template<class DataT, class Allocator>
    void process_fake(Allocator& allocator, uint32_t period_microsec = 250, ProcessCallback callback = nullptr);

    uint32_t get_pdo_counter() {
        return pdo_counter_;
    }

    template<typename... DataTypes>
    bool pdo_write(uint32_t timeout_ticks, address::Offset offset, DataTypes&... data) {
        custom_promise::WritePromise<DataTypes...> promise{offset, data...};
        auto future = promise.get_future();

        mutex_write_.lock();

        queue_write_ + promise;

        mutex_write_.unlock();

        future.get(timeout_ticks);

        return future.is_ready();
    }

    template<typename... DataTypes>
    bool pdo_read(uint32_t timeout_ticks, address::Offset offset, DataTypes&... data) {
        custom_promise::ReadPromise<DataTypes...> promise{offset, data...};
        auto future = promise.get_future();

        mutex_read_.lock();

        queue_read_ + promise;

        mutex_read_.unlock();

        future.get(timeout_ticks);

        return future.is_ready();
    }


    void pdo_write_async_static(custom_promise::IPromise& promise) {
        mutex_write_.lock();
        static_queue_write_ + promise;
        mutex_write_.unlock();
    }

    void pdo_read_async_static(custom_promise::IPromise& promise) {
        mutex_read_.lock();
        static_queue_read_ + promise;
        mutex_read_.unlock();
    }

    void pdo_write_async(custom_promise::IPromise& promise) {
        mutex_write_.lock();
        queue_write_ + promise;
        mutex_write_.unlock();
    }

    void pdo_read_async(custom_promise::IPromise& promise) {
        mutex_read_.lock();
        queue_read_ + promise;
        mutex_read_.unlock();
    }

private:
    ecat_buffer::EthEcatBuffer& ecat_buffer_;

    std::vector<EcatPdoFMMUSlave> pdo_fmmu_slaves_;

    Mutex mutex_write_;
    Mutex mutex_read_;

    queue::Queue<custom_promise::IPromise> static_queue_write_;
    queue::Queue<custom_promise::IPromise> static_queue_read_;

    queue::Queue<custom_promise::IPromise> queue_write_;
    queue::Queue<custom_promise::IPromise> queue_read_;

    ProcessCallback callback_{nullptr};

    uint32_t pdo_counter_{0};

    template<class DataT, class Allocator>
    bool wait_op(std::vector<DataT, Allocator>& pdo_read, std::vector<DataT, Allocator>& pdo_write);

    void process_write_queue(uint8_t* process_data, uint32_t len);
    void process_read_queue(uint8_t* process_data, uint32_t len);

    template<typename... DataTypes>
    bool write(address::Offset offset, DataTypes&... data) {
        using TDatagram = datagram::EcatDatagram<command::LWR, DataTypes...>;

        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
        ecat_buffer::FMMUGlobalProperties& fmmu_global_properties = ecat_buffer_.get_fmmu_global_properties();
        address::Logical logical = fmmu_global_properties.logical_start_address + offset;
        const datagram::TEcatWkc expected_wkc = pdo_fmmu_slaves_.size();
        TDatagram datagram{ {{logical}}, expected_wkc, data... };

        return telegram.transfer(datagram);
    }

    template<typename... DataTypes>
    bool read(address::Offset offset, DataTypes&... data) {
        using TDatagram = datagram::EcatDatagram<command::LRD, DataTypes...>;

        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
        ecat_buffer::FMMUGlobalProperties& fmmu_global_properties = ecat_buffer_.get_fmmu_global_properties();
        address::Logical logical = fmmu_global_properties.logical_start_address + fmmu_global_properties.logical_full_length_write + offset;
        const datagram::TEcatWkc expected_wkc = pdo_fmmu_slaves_.size();
        TDatagram datagram{ {{logical}}, expected_wkc, data... };

        return telegram.transfer(datagram);
    }

    template<typename... DataTypes>
    bool read_write(address::Offset offset_read, custom_tuple<DataTypes&...> data_read, address::Offset offset_write, custom_tuple<DataTypes&...> data_write) {
        using TDatagramRead = datagram::EcatDatagram<command::LRD, DataTypes...>;
        using TDatagramWrite = datagram::EcatDatagram<command::LWR, DataTypes...>;

        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();
        ecat_buffer::FMMUGlobalProperties& fmmu_global_properties = ecat_buffer_.get_fmmu_global_properties();
        const datagram::TEcatWkc expected_wkc = pdo_fmmu_slaves_.size();

        address::Logical logical_read = fmmu_global_properties.logical_start_address + fmmu_global_properties.logical_full_length_write + offset_read;
        TDatagramRead datagram_read{ {{logical_read}}, expected_wkc, data_read };

        address::Logical logical_write = fmmu_global_properties.logical_start_address + offset_write;
        TDatagramWrite datagram_write{ {{logical_write}}, expected_wkc, data_write };

        auto queue = datagram_read + datagram_write;

        return telegram.transfer(queue);
    }

};

template<class DataT, class Allocator>
bool EthEcatPdoFMMU::wait_op(std::vector<DataT, Allocator>& pdo_read, std::vector<DataT, Allocator>& pdo_write) {
    free_rtos::Semaphore& init_sem = ecat_buffer_.get_ecat().get_init_sem();
    free_rtos::Semaphore& process_sem = ecat_buffer_.get_ecat().get_process_sem();

    process_sem.pend();

    uint32_t logical_full_length_write = ecat_buffer_.get_fmmu_global_properties().logical_full_length_write + sizeof(DataT) - 1;
    uint32_t logical_full_length_read = ecat_buffer_.get_fmmu_global_properties().logical_full_length_read + sizeof(DataT) - 1;
    custom_tuple< std::vector<DataT, Allocator>& > data_tuple_read{pdo_read};
    custom_tuple< std::vector<DataT, Allocator>& > data_tuple_write{pdo_write};
    bool status;

    pdo_read.reserve(logical_full_length_read/sizeof(DataT));
    pdo_read.resize(logical_full_length_read/sizeof(DataT), 0);

    pdo_write.reserve(logical_full_length_write/sizeof(DataT));
    pdo_write.resize(logical_full_length_write/sizeof(DataT), 0);

    status = read(0, pdo_read); // read to pdo_write !

    if(status != true) {
        return status;
    }

    status = write(0, pdo_write);

    if(status != true) {
        return status;
    }

    init_sem.post();
    process_sem.pend();

    for(uint32_t i = 0; i < 250; i++) {
        status = read_write(0, data_tuple_read, 0, data_tuple_write);

        if(status != true) {
            return status;
        }
    }

    init_sem.post();
    process_sem.pend();

    return status;
}

template<class DataT, class Allocator>
void EthEcatPdoFMMU::process(Allocator& allocator) {
    std::vector<DataT, Allocator> pdo_read{allocator};
    std::vector<DataT, Allocator> pdo_write{allocator};
    bool status;

    status = wait_op(pdo_read, pdo_write);

    if(status != true) {
        return;
    }

    custom_tuple< std::vector<DataT, Allocator>& > data_tuple_read{pdo_read};
    custom_tuple< std::vector<DataT, Allocator>& > data_tuple_write{pdo_write};

    status = read(0, pdo_read); // read to pdo_write !

    if(status != true) {
        return;
    }

    status = write(0, pdo_write);

    if(status != true) {
        return;
    }

    while(1) {
        status = read_write(0, data_tuple_read, 0, data_tuple_write);

        if(status != true) {
            return;
        }

        #if PROFILE_LEVEL > 0
        {
            #warning ( "Profiling for RPDO subscribers is enabled !" )

            PROFILE(process_read_queue, PROFILE_SAMPLES, 0);

            process_read_queue(pdo_read.data(), pdo_read.size());
        }
        #else
        process_read_queue(pdo_read.data(), pdo_read.size());
        #endif

        #if PROFILE_LEVEL > 0
        {
            #warning ( "Profiling for process callback is enabled !" )

            PROFILE(callback_, PROFILE_SAMPLES, 0);

            if(callback_ != nullptr) {
                callback_();
            }
        }
        #else
        if(callback_ != nullptr) {
            callback_();
        }
        #endif

        #if PROFILE_LEVEL > 0
        {
            #warning ( "Profiling for TPDO subscribers is enabled !" )

            PROFILE(process_write_queue, PROFILE_SAMPLES, 0);

            process_write_queue(pdo_write.data(), pdo_write.size());
        }
        #else
        process_write_queue(pdo_write.data(), pdo_write.size());
        #endif

        pdo_counter_++;
    }
}

template<class DataT, class Allocator>
void EthEcatPdoFMMU::process_fake(Allocator& allocator, uint32_t period_microsec, ecat_pdo_fmmu::ProcessCallback callback) {
    std::vector<DataT, Allocator> pdo_read(110, 0, allocator);
    std::vector<DataT, Allocator> pdo_write(110, 0, allocator);
    free_rtos::Timer& ecat_timer = ecat_buffer_.get_ecat().get_ecat_timer();

    Timer::Settings ecat_tmr_sett = {
                 .input_clk_Hz = 25000000, // 25MHz
                 .base_address = 0x2400000u, // memory mapping,
                 .clock_src_mux_addr = 0x430081B0u, // sysconfig
                 .int_num = 152u,                   // sysconfig
                 .int_priority = 4,                 // sysconfig
                 .period_us = period_microsec /// microsec
              };

    ecat_timer.Init(ecat_tmr_sett);
    ecat_timer.Start();

    while(1) {
        ecat_timer.Wait();

        #if PROFILE_LEVEL > 0
        {
            #warning ( "Profiling for RPDO subscribers is enabled !" )

            PROFILE(process_read_queue, PROFILE_SAMPLES, 0);

            process_read_queue(pdo_read.data(), pdo_read.size());
        }
        #else
        process_read_queue(pdo_read.data(), pdo_read.size());
        #endif

        #if PROFILE_LEVEL > 0
        {
            #warning ( "Profiling for process callback is enabled !" )

            PROFILE(callback_, PROFILE_SAMPLES, 0);

            if(callback_ != nullptr) {
                callback_();
            }
        }
        #else
        if(callback != nullptr) {
            callback();
        }
        #endif

        #if PROFILE_LEVEL > 0
        {
            #warning ( "Profiling for TPDO subscribers is enabled !" )

            PROFILE(process_write_queue, PROFILE_SAMPLES, 0);

            process_write_queue(pdo_write.data(), pdo_write.size());
        }
        #else
        process_write_queue(pdo_write.data(), pdo_write.size());
        #endif

        pdo_counter_++;
    }
}

} // namespace ecat_pdo_fmmu

} // namespace free_rtos

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_PDO_FMMU_HPP_ */
