/*
 * eth_ecat_sdo_mailbox.hpp
 *
 *  Created on: May 16, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_SDO_MAILBOX_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_SDO_MAILBOX_HPP_

#include "free_rtos/ethernet_industry/eth_ecat_buffer.hpp"

namespace free_rtos {

namespace ecat_sdo_mailbox {

enum class SDOReqCommandSpecifier : std::uint16_t {
    DOWNLOAD_SEGMENT = 0x0,
    DOWNLOAD,
    UPLOAD,
    UPLOAD_SEGMENT,
    ABORT_TRANSFER
};

enum class TransferType : std::uint16_t {
    NORMAL = 0x0,
    EXPEDITED
};

enum class SDORespCommandSpecifier : std::uint16_t {
    UPLOAD_SEGMENT = 0x0,
    DOWNLOAD_SEGMENT,
    UPLOAD,
    DOWNLOAD,
    ABORT_TRANSFER
};

enum class Service : uint16_t {
    EMERGENCY = 0x1,
    SDO_REQUEST,
    SDO_RESPONSE,
    TX_PDO,
    RX_PDO,
    TX_PDO_REMOTE_REQUEST,
    RX_PDO_REMOTE_REQUEST,
    SDO_INFORMATION
};

struct CompleteSize {
    uint32_t value;
} __attribute__ ((packed));

struct CommandProp {
    std::uint8_t size : 1;
    std::uint8_t transfer_type : 1;
    std::uint8_t data_set_size : 2;
    std::uint8_t complete_access : 1;
    std::uint8_t command_spec : 3;
} __attribute__ ((packed));

struct CoEHeader {
    std::uint8_t number;
    std::uint8_t reserved : 4;
    std::uint8_t service : 4;
} __attribute__ ((packed));

struct CoEElements {
    CoEHeader coe_header;
    CommandProp command_specifier;
    std::uint16_t index;
    std::uint8_t subindex;
} __attribute__ ((packed));

enum class ProtocolType : std::uint16_t {
    /** Beckhoff: AoE ADS over EtherCAT */
    VENDOR_SPECIFIC = 1,
    /** Ethernet over EtherCAT */
    EoE,
    /** CanOpen Over EtherCAT */
    CoE,
    /** File access over EtherCAT */
    FoE,
    /** Servo profile over EtherCAT */
    SoE,
    /** Vendor specific (VoE) */
    VoE = 0xF
};

struct MailboxHeader {
    uint16_t length;
    uint16_t address;
    uint16_t channel : 6;
    uint16_t priority : 2;
    ProtocolType type : 4;
    uint16_t cnt : 3;
    uint16_t reserved : 1;
} __attribute__ ((packed));

} // namespace ecat_sdo_mailbox

// Специализация шаблона для распаковки протокола CoE
// наследоваться от each_tuple_element< FunctorT, custom_tuple<ecat_sdo_mailbox::CompleteSize&, TailT...> > не обязательно
// т.к. используются только статические методы
template<typename FunctorT, typename... TailT>
struct each_tuple_element< FunctorT, custom_tuple<ecat_sdo_mailbox::CoEElements&, ecat_sdo_mailbox::CompleteSize&, TailT...> > {
    using TTuple = custom_tuple<ecat_sdo_mailbox::CoEElements&, ecat_sdo_mailbox::CompleteSize&, TailT...>;
    using TBase = each_tuple_element<FunctorT, typename TTuple::TBase>;
    using TBaseBase = each_tuple_element<FunctorT, typename TTuple::TBase::TBase>;

    static void for_each(FunctorT& functor, TTuple& t) {
        functor(t.head_);

        if(t.head_.command_specifier.transfer_type == static_cast<uint8_t>(ecat_sdo_mailbox::TransferType::NORMAL)) {
            TBase::for_each(functor, t);
        }

        if(t.head_.command_specifier.transfer_type == static_cast<uint8_t>(ecat_sdo_mailbox::TransferType::EXPEDITED)) {
            TBaseBase::for_each(functor, t);

            Padding padding{sizeof(typename TTuple::TBase::THead)};
            functor(padding);
        }
    }

    // Не используется, но пускай будет
    static void for_each_reverse(FunctorT& functor, TTuple& t) {
        if(t.head_.command_specifier.transfer_type == static_cast<uint8_t>(ecat_sdo_mailbox::TransferType::NORMAL)) {
            TBase::for_each_reverse(functor, t);
        }

        if(t.head_.command_specifier.transfer_type == static_cast<uint8_t>(ecat_sdo_mailbox::TransferType::EXPEDITED)) {
            Padding padding{sizeof(typename TTuple::TBase::THead)};
            functor(padding);

            TBaseBase::for_each_reverse(functor, t);
        }

        functor(t.head_);
    }
};

namespace ecat_sdo_mailbox {

class EcatSdoMailboxSlave {
public:
    EcatSdoMailboxSlave(ecat_buffer::EcatBufferSlave& mailbox_slave)
    : buffer_slave_(mailbox_slave) { }

    template<typename TypeT>
    bool wait_available(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc) {
        using TCommand = command::EcatCommand<TypeT, command::RD>;

        auto slave_address = buffer_slave_.get_slave().get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        uint8_t sm_status{0x00};
        datagram::EcatDatagram<TCommand, uint8_t> datagram{ {{slave_address, buffer_regs[MailboxesRegs::AVAILABLE]}}, expected_wkc, sm_status };
        bool status;

        do {
            status = telegram.transfer(datagram);
        } while((status == true) && ((sm_status & 0x08) == 0));

        return status;
    }

    template<typename TypeT>
    bool wait_empty(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc) {
        using TCommand = command::EcatCommand<TypeT, command::RD>;

        auto slave_address = buffer_slave_.get_slave().get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        uint8_t sm_status{0x00};
        datagram::EcatDatagram<TCommand, uint8_t> datagram{ {{slave_address, buffer_regs[MailboxesRegs::EMPTY]}}, expected_wkc, sm_status };
        bool status;

        do {
            status = telegram.transfer(datagram);
        } while((status == true) && ((sm_status & 0x08) != 0));

        return status;
    }

    template<typename TypeT>
    bool empty(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc) {
        using TCommand = command::EcatCommand<TypeT, command::RD>;

        auto slave_address = buffer_slave_.get_slave().get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        uint8_t sm_status{0x00};
        datagram::EcatDatagram<TCommand, uint8_t> datagram{ {{slave_address, buffer_regs[MailboxesRegs::AVAILABLE]}}, expected_wkc, sm_status };

        return telegram.transfer(datagram);
    }

    void increment_counter() {
        counter_++;

        if(counter_ > 7) {
            counter_ = 1;
        }
    }

    template<typename TypeT, typename... DataTypes>
    bool send_data(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc, uint16_t channel, uint16_t priority, ProtocolType type, DataTypes&... data) {
        using TCommand = command::EcatCommand<TypeT, command::WR>;
        using TDatagram = datagram::EcatDatagram<TCommand, MailboxHeader, DataTypes... , Padding>;

        EcatSlave& slave = buffer_slave_.get_slave();
        auto slave_address = slave.get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        MailboxHeader header{
            .length = custom_tuple<DataTypes...>::size,
            .address = 0, // slave.get_slave_address<command::FP>()
            .channel = channel,
            .priority = priority,
            .type = type,
            .cnt = static_cast<uint16_t>(counter_)
        };
        Padding padding{buffer_slave_.get_buffer_properties_write().length - sizeof(MailboxHeader) - custom_tuple<DataTypes...>::size};
        TDatagram datagram{ {{slave_address, buffer_regs[MailboxesRegs::WRITE]}}, expected_wkc, header, data... , padding};
        bool status;

        status = telegram.transfer(datagram);

        if(status != true) {
            return status;
        }

        increment_counter();

        return status;
    }

    template<typename TypeT, typename... DataTypes>
    bool receive_data(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc, DataTypes&... data) {
        using TCommand = command::EcatCommand<TypeT, command::RD>;
        using TDatagram = datagram::EcatDatagram<TCommand, MailboxHeader, DataTypes... , Padding>;

        auto slave_address = buffer_slave_.get_slave().get_slave_address<TypeT>();
        std::array<address::Offset, 4>& buffer_regs = buffer_slave_.get_buffer_regs();
        MailboxHeader header;
        Padding padding{buffer_slave_.get_buffer_properties_read().length - sizeof(MailboxHeader) - custom_tuple<DataTypes...>::size};
        TDatagram datagram{ {{slave_address, buffer_regs[MailboxesRegs::READ]}}, expected_wkc, header, data... , padding};

        return telegram.transfer(datagram);

        //DebugP_log("header.length = %d\r\n", header.length);
        //DebugP_log("header.address = %d\r\n", header.address);
        //DebugP_log("header.channel = %d\r\n", header.channel);
        //DebugP_log("header.priority = %d\r\n", header.priority);
        //DebugP_log("header.type = %d\r\n", header.type);
        //DebugP_log("header.cnt = %d\r\n", header.cnt);
    }

    template<typename TypeT, typename... DataTypes>
    bool receive(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc, DataTypes&... data) {
        if(wait_available<TypeT>(telegram, expected_wkc) != true) return false;
        if(receive_data<TypeT, DataTypes...>(telegram, expected_wkc, data...) != true) return false;

        return true;
    }

    template<typename TypeT, typename... DataTypes>
    bool send(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc, uint16_t channel, uint16_t priority, ProtocolType type, DataTypes&... data) {
        if(empty<TypeT>(telegram, expected_wkc) != true) return false;
        if(wait_empty<TypeT>(telegram, expected_wkc) != true) return false;
        if(send_data<TypeT, DataTypes...>(telegram, expected_wkc, channel, priority, type, data...) != true) return false;
        if(wait_empty<TypeT>(telegram, expected_wkc) != true) return false;

        return true;
    }

    template<typename TypeT, typename... DataTypes>
    bool sdo_write(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc, uint16_t index, uint8_t subindex, DataTypes&... data) {
        CoEElements elements{
            .coe_header = {
                .number = 0x00,
                .service = static_cast<uint8_t>(Service::SDO_REQUEST) },
            .command_specifier = {
                .size = 1,
                .transfer_type = static_cast<uint8_t>(TransferType::NORMAL),
                .data_set_size = 0,
                .complete_access = 0,
                .command_spec = static_cast<uint8_t>(SDOReqCommandSpecifier::DOWNLOAD)},
            .index = index,
            .subindex = subindex };
        CompleteSize complete_size{custom_tuple<DataTypes...>::size};
        bool status;

        status = send<TypeT, CoEElements, CompleteSize>(telegram, expected_wkc, 0, 0, ProtocolType::CoE, elements, complete_size, data...);

        if(status != true) {
            return status;
        }

        status = receive<TypeT, CoEElements, CompleteSize, DataTypes...>(telegram, expected_wkc, elements, complete_size);

        if(status != true) {
            return status;
        }

        if( (elements.coe_header.service != static_cast<uint8_t>(Service::SDO_RESPONSE)) ||
            (elements.command_specifier.command_spec != static_cast<uint8_t>(SDOReqCommandSpecifier::UPLOAD)) ) {

            DebugP_log((char*)"CoE error: = 0x%04x\r\n", complete_size.value); // 0x601004 - The object cannot be accessed via complete access
        }

        //DebugP_log("elements.coe_header.number = %d\r\n", elements.coe_header.number);
        //DebugP_log("elements.coe_header.service = %d\r\n", elements.coe_header.service);
        //DebugP_log("elements.command_specifier.size = %d\r\n", elements.command_specifier.size);
        //DebugP_log("elements.command_specifier.transfer_type = %d\r\n", elements.command_specifier.transfer_type);
        //DebugP_log("elements.command_specifier.data_set_size = %d\r\n", elements.command_specifier.data_set_size);
        //DebugP_log("elements.command_specifier.complete_access = %d\r\n", elements.command_specifier.complete_access);
        //DebugP_log("elements.command_specifier.command_spec = %d\r\n", elements.command_specifier.command_spec);
        //DebugP_log("elements.index = %d\r\n", elements.index);
        //DebugP_log("elements.subindex = %d\r\n", elements.subindex);
        //DebugP_log("complete_size = %d\r\n", complete_size);

        return status;
    }

    template<typename TypeT, typename... DataTypes>
    bool sdo_read(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc, uint16_t index, uint8_t subindex, DataTypes&... data) {
        CoEElements elements{
            .coe_header = {
                .number = 0x00,
                .service = static_cast<uint8_t>(Service::SDO_REQUEST) },
            .command_specifier = {
                .size = 1,
                .transfer_type = static_cast<uint8_t>(TransferType::NORMAL),
                .data_set_size = 0,
                .complete_access = 0,
                .command_spec = static_cast<uint8_t>(SDOReqCommandSpecifier::UPLOAD)},
            .index = index,
            .subindex = subindex };
        CompleteSize complete_size{0x00000000};
        bool status;

        status = send<TypeT, CoEElements, CompleteSize>(telegram, expected_wkc, 0, 0, ProtocolType::CoE, elements, complete_size);

        if(status != true) {
            return status;
        }

        status = receive<TypeT, CoEElements, CompleteSize, DataTypes...>(telegram, expected_wkc, elements, complete_size, data...);

        if(status != true) {
            return status;
        }

        if( (elements.coe_header.service != static_cast<uint8_t>(Service::SDO_RESPONSE)) ||
            (elements.command_specifier.command_spec != static_cast<uint8_t>(SDOReqCommandSpecifier::UPLOAD)) ) {

            DebugP_log((char*)"CoE error: = 0x%04x\r\n", complete_size.value); // 0x601004 - The object cannot be accessed via complete access
        }

        //DebugP_log("elements.coe_header.number = %d\r\n", elements.coe_header.number);
        //DebugP_log("elements.coe_header.service = %d\r\n", elements.coe_header.service);
        //DebugP_log("elements.command_specifier.size = %d\r\n", elements.command_specifier.size);
        //DebugP_log("elements.command_specifier.transfer_type = %d\r\n", elements.command_specifier.transfer_type);
        //DebugP_log("elements.command_specifier.data_set_size = %d\r\n", elements.command_specifier.data_set_size);
        //DebugP_log("elements.command_specifier.complete_access = %d\r\n", elements.command_specifier.complete_access);
        //DebugP_log("elements.command_specifier.command_spec = %d\r\n", elements.command_specifier.command_spec);
        //DebugP_log("elements.index = %d\r\n", elements.index);
        //DebugP_log("elements.subindex = %d\r\n", elements.subindex);
        //DebugP_log("complete_size = %d\r\n", complete_size);

        return status;
    }

    template<typename TypeT>
    bool pdo_map_read(telegram::EcatTelegram& telegram, const datagram::TEcatWkc expected_wkc, ecat_buffer::PDOMap& pdo_map, uint16_t pdo_map_index) {
        uint16_t pdo_data_size{0x0000}; // Size is in bits !
        uint8_t pdo_block_count{0x00};
        bool status;

        pdo_map.block_index_map.emplace(pdo_map_index, std::map<uint8_t, uint16_t>{});

        status = sdo_read<TypeT, uint8_t>(telegram, expected_wkc, pdo_map_index, 0x00, pdo_block_count);

        if(status != true) {
            return status;
        }

        pdo_map.block_index_map[pdo_map_index].emplace(0, pdo_block_count);

        DebugP_log((char*)"pdo_block_count = 0x%01x\r\n", pdo_block_count);

        for(uint8_t pdo_map_subindex = 1; pdo_map_subindex < (pdo_block_count + 1); pdo_map_subindex++) {
            uint16_t pdo_block_index{0x0000};

            status = sdo_read<TypeT, uint16_t>(telegram, expected_wkc, pdo_map_index, pdo_map_subindex, pdo_block_index);

            if(status != true) {
                return status;
            }

            pdo_map.block_index_map[pdo_map_index].emplace(pdo_map_subindex, pdo_block_index);

            //DebugP_log((char*)"pdo_block_index = 0x02%x\r\n", pdo_block_index);

            ecat_buffer::PDODescriptor tmp;
            uint8_t& pdo_block_object_count = tmp.size;

            pdo_map.object_descriptor_map.emplace(pdo_block_index, std::map<uint8_t, ecat_buffer::PDODescriptor>{});

            status = sdo_read<TypeT, uint8_t>(telegram, expected_wkc, pdo_block_index, 0, pdo_block_object_count);

            if(status != true) {
                return status;
            }

            pdo_map.object_descriptor_map[pdo_block_index].emplace(0, tmp);

            DebugP_log((char*)"pdo_block_object_count = 0x%01x\r\n", pdo_block_object_count);

            for(uint8_t pdo_block_subindex = 1; pdo_block_subindex < (pdo_block_object_count + 1); pdo_block_subindex++) {
                ecat_buffer::PDODescriptor descriptor;

                status = sdo_read<TypeT, ecat_buffer::PDODescriptor>(telegram, expected_wkc, pdo_block_index, pdo_block_subindex, descriptor);

                if(status != true) {
                    return status;
                }

                pdo_data_size += descriptor.size;
                pdo_map.object_descriptor_map[pdo_block_index].emplace(pdo_block_subindex, descriptor);

                //DebugP_log("descriptor.size = 0x%01x\r\n", descriptor.size);
                //DebugP_log("descriptor.subindex = 0x%01x\r\n", descriptor.subindex);
                //DebugP_log("descriptor.index = 0x%02x\r\n", descriptor.index);
            }
        }

        pdo_map.data_size_map.emplace(pdo_map_index, pdo_data_size/8);

        DebugP_log((char*)"pdo_data_size = %d\r\n", pdo_data_size/8);

        return status;
    }

private:
    ecat_buffer::EcatBufferSlave& buffer_slave_;

    uint8_t counter_{0x00};
};

class EthEcatSdoMailbox {
public:
    EthEcatSdoMailbox(ecat_buffer::EthEcatBuffer& ecat_buffer): ecat_buffer_{ecat_buffer} { }

    void init();
    bool read_pdo_map(uint16_t pdo_map_rx_index, uint16_t pdo_map_tx_index);

    std::vector<ecat_buffer::PDOMap>& get_pdo_map() {
        return pdo_map_;
    }

    template<typename TypeT, typename... DataTypes>
    CompleteSize sdo_write(const datagram::TEcatWkc expected_wkc, size_t slave_index, uint16_t index, uint8_t subindex, DataTypes&... data) {
        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();

        return sdo_mailbox_slaves_[slave_index].sdo_write<TypeT, DataTypes...>(telegram, expected_wkc, index, subindex, data...);
    }

    template<typename TypeT, typename... DataTypes>
    CompleteSize sdo_read(const datagram::TEcatWkc expected_wkc, size_t slave_index, uint16_t index, uint8_t subindex, DataTypes&... data) {
        telegram::EcatTelegram& telegram = ecat_buffer_.get_ecat().get_telegram();

        return sdo_mailbox_slaves_[slave_index].sdo_read<TypeT, DataTypes...>(telegram, expected_wkc, index, subindex, data...);
    }

private:
    ecat_buffer::EthEcatBuffer& ecat_buffer_;

    std::vector<EcatSdoMailboxSlave> sdo_mailbox_slaves_;

    std::vector<ecat_buffer::PDOMap> pdo_map_;
};

} // namespace ecat_sdo_mailbox

} // namespace free_rtos

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_SDO_MAILBOX_HPP_ */
