/*
 * eth_ecat_custom_promise.hpp
 *
 *  Created on: Jun 1, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_CUSTOM_PROMISE_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_CUSTOM_PROMISE_HPP_

#include <cstdint>

#include "free_rtos/semaphore/semaphore.hpp"

#include "free_rtos/ethernet_industry/eth_ecat_queue.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_datagram.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_command.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_custom_tuple.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_packer.hpp"

namespace free_rtos {

namespace custom_promise {

template<typename... DataTypes>
class Future {
public:
    Future(DataTypes&... data)
    : data_tuple_{data...} { }

    bool is_ready() {
        return ready_;
    }

    custom_tuple<DataTypes&...> get(uint32_t timeout_ticks = SystemP_WAIT_FOREVER) {
        int32_t sts;

        ready_ = false;
        sts = sem_.pend(timeout_ticks);

        if(sts == SystemP_SUCCESS) {
            ready_ = true;
        }

        return data_tuple_;
    }

    void pack(uint8_t* raw) {
        PackFunctor functor{raw, false};

        for_each(data_tuple_, functor);

        sem_.post();
    }

    void unpack(uint8_t* raw) {
        UnpackFunctor functor{raw, false};

        for_each(data_tuple_, functor);

        sem_.post();
    }

private:
    custom_tuple<DataTypes&...> data_tuple_;
    free_rtos::Semaphore sem_;
    bool ready_{false};
};

class IPromise : protected queue::QueueEntity {
    friend class queue::Queue<IPromise>;

public:
    IPromise(address::Offset offset = 0)
    : offset_{offset} { }

    queue::Queue<IPromise> operator+(IPromise &next) {
        return queue::Queue<IPromise>{*this, next};
    }

    virtual void set_value(uint8_t* process_data, uint32_t len) = 0;

protected:
    address::Offset offset_{0};

private:

};

template<typename... DataTypes>
class WritePromise : public IPromise {
public:
    WritePromise(address::Offset offset, DataTypes&... data)
    : IPromise{offset}
    , future_{data...} { }

    Future<DataTypes...>& get_future() {
        return future_;
    }

    virtual void set_value(uint8_t* process_data, uint32_t len) override {
        future_.pack(process_data + offset_);
    }

private:
    Future<DataTypes...> future_;
};

template<typename... DataTypes>
class ReadPromise : public IPromise {
public:
    ReadPromise(address::Offset offset, DataTypes&... data)
    : IPromise{offset}
    , future_{data...} { }

    Future<DataTypes...>& get_future() {
        return future_;
    }

    virtual void set_value(uint8_t* process_data, uint32_t len) override {
        future_.unpack(process_data + offset_);
    }

private:
    Future<DataTypes...> future_;
};

/*
Объект FunctorT реализует пользователь под собственные нужды и должен иметь вид
либо функтора, либо лямбды.

// Пример пользовательского функтора
class UserFunctor {
public:
    // В конструктор можно передать переменные из пользовательского контекста
    UserFunctor(address::Offset offset, std::vector<uint8_t>& data)
    : offset_{offset}
    , data_{data} { }

    bool get_ready() {
        return ready_;
    }

    void set_ready(bool ready) {
        ready_ = ready;
    }

    void operator()(uint8_t* raw, uint32_t len) {
        // Обработка коллбека пользователем
        size_t size = data_.size() * sizeof(uint8_t);

        memcpy(data_.data(), raw + offset_, size);

        ready_ = true;

        // DebugP_log((char*)"Запрос обработан !\r\n");
    }

private:
    // Переменные из пользовательского контекста
    address::Offset offset_{0};
    std::vector<uint8_t>& data_;

    bool ready_{false};
};

// Пример пользовательской лямбды
auto lambda = [&](uint8_t* raw, uint32_t len) {
    // Обработка коллбека пользователем
    size_t size = process_data.size() * sizeof(uint8_t);
    address::Offset offset{0};

    memcpy(process_data.data(), raw + offset, size);

    ready = true;

    // DebugP_log((char*)"Запрос обработан !\r\n");
};

*/

template<typename FunctorT>
class FunctorPromise : public IPromise {
public:
    FunctorPromise(FunctorT& functor)
    : IPromise{0}
    , functor_(functor) { }

    virtual void set_value(uint8_t* process_data, uint32_t len) override {
        functor_(process_data, len);
    }

private:
    FunctorT& functor_;
};

} // namespace custom_promise

} // namespace free_rtos

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_COE_ETH_ECAT_CUSTOM_PROMISE_HPP_ */
