/*
 * eth_ecat_datagram.hpp
 *
 *  Created on: May 2, 2023
 *      Author: algin
 */

#ifndef FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_
#define FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_

#include <cstdint>

#include "free_rtos/ethernet_industry/ethercattype.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_types.h"
#include "free_rtos/ethernet_industry/eth_ecat_queue.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_custom_tuple.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_packer.hpp"
#include "free_rtos/ethernet_industry/eth_ecat_command.hpp"

namespace free_rtos {

namespace datagram {

using TEcatWkc = uint16_t;

class IEcatDatagram : public queue::QueueEntity {
    friend class queue::Queue<IEcatDatagram>;

public:
    IEcatDatagram()
    : expected_wkc_{0x0001}
    { }

    IEcatDatagram(TEcatWkc expected_wkc)
    : expected_wkc_{expected_wkc}
    { }

    virtual ~IEcatDatagram() { };

    queue::Queue<IEcatDatagram> operator+(IEcatDatagram &next) {
        //DebugP_log((char*)"Coupling first pair\r\n");

        return queue::Queue<IEcatDatagram>{*this, next};
    }

    virtual uint8_t* pack(uint8_t *raw, uint8_t idx) = 0;
    virtual uint8_t* unpack(uint8_t *raw, uint8_t idx) = 0;

    virtual uint8_t* pack_all(uint8_t *raw, uint8_t idx) = 0;
    virtual uint8_t* unpack_all(uint8_t *raw, uint8_t idx) = 0;

    TEcatDgHeader& get_header() {
        return header_;
    }

    TEcatWkc get_wkc() {
        return wkc_;
    }

    void reset_wkc() {
        wkc_ = 0x0000;
    }

    bool as_expected() {
        return (wkc_ >= expected_wkc_);
    }

    TEcatWkc get_all_wkc() {
        IEcatDatagram* next = this;
        TEcatWkc wkc = 0x0000;

        do {
            wkc += next->get_wkc();

            next = static_cast<IEcatDatagram*>(next->get_next());
        } while(next != nullptr);

        return wkc;
    }

    void reset_all_wkc() {
        IEcatDatagram* next = this;

        do {
            next->reset_wkc();

            next = static_cast<IEcatDatagram*>(next->get_next());
        } while(next != nullptr);
    }

    bool all_as_expected() {
        IEcatDatagram* next = this;
        bool as_expected = true;

        do {
            as_expected = (as_expected && next->as_expected());

            next = static_cast<IEcatDatagram*>(next->get_next());
        } while(next != nullptr);

        return as_expected;
    }

protected:
    TEcatDgHeader header_;
    TEcatWkc wkc_{0x0000};
    const TEcatWkc expected_wkc_{0x0001};

private:

};

/*
 * EtherCAT Datagram
 *
 * Первый параметр шаблона должен быть class EcatCommand
 *
 * Возможные значения первого параметра шаблона:
 *
 * EcatDatagram< BWR, DataType >
 * или
 * EcatDatagram< EcatCommand< B, WR >, DataType >
 *
 * EcatDatagram< LWR, DataType >
 * или
 * EcatDatagram< EcatCommand< L, WR >, DataType >
 *
 * Пример:
 *
 * datagram::EcatDatagram<command::BWR, DataType > a{ {{broadcast, address}}, expected_wkc, data };
 * datagram::EcatDatagram<command::LWR, DataType > b{ {{logic, address}}, expected_wkc, data };
 *
 * auto queue = a + b;
 *
 */
template<typename CommandT, typename... DataTypes>
class EcatDatagram : public IEcatDatagram {
    static_assert(std::is_base_of<command::EcatCommandBase, CommandT>::value == true, "CommandT should be derived from ECatCommandBase");

public:
    EcatDatagram(CommandT&& command, const TEcatWkc expected_wkc, DataTypes&... data)
    : IEcatDatagram{expected_wkc}
    , command_{command}
    , data_tuple_{data...} { }

    EcatDatagram(CommandT&& command, const TEcatWkc expected_wkc, custom_tuple<DataTypes&...> data_tuple)
    : IEcatDatagram{expected_wkc}
    , command_{command}
    , data_tuple_{data_tuple} { }

    ~EcatDatagram() { }

    virtual uint8_t* pack(uint8_t *raw, uint8_t idx) override {
        if((as_expected() == true) && (command_.get_cmd() != EC_CMD_NOP)) {
            return raw;
        }

        return pack_header(raw, idx);
    }

    virtual uint8_t* unpack(uint8_t *raw, uint8_t idx) override {
        if((as_expected() == true) && (command_.get_cmd() != EC_CMD_NOP)) {
            return raw;
        }

        return unpack_header(raw, idx);
    }

    virtual uint8_t* pack_all(uint8_t *raw, uint8_t idx) override {
        IEcatDatagram* next = this;

        do {
            raw = next->pack(raw, idx);

            next = static_cast<IEcatDatagram*>(next->get_next());
        } while(next != nullptr);

        return raw;
    }

    virtual uint8_t* unpack_all(uint8_t *raw, uint8_t idx) override {
        IEcatDatagram* next = this;

        do {
            raw = next->unpack(raw, idx);

            if(raw == nullptr) {
                break;
            }

            next = static_cast<IEcatDatagram*>(next->get_next());
        } while(next != nullptr);

        return raw;
    }

private:
    CommandT command_;
    custom_tuple<DataTypes&...> data_tuple_;

    uint8_t* pack_wkc(uint8_t *raw) {
        TEcatWkc *wkc = new(raw) TEcatWkc{0x0000};

        (void)wkc;

        return raw + sizeof(TEcatWkc);
    }

    uint8_t* pack_data(uint8_t *raw) {
        PackFunctor functor{raw, command_.get_skip_pack()};

        for_each(data_tuple_, functor);

        return functor.raw;
    }

    uint8_t* pack_header(uint8_t *raw, uint8_t idx) {
        uint8_t *data_raw = raw + sizeof(TEcatDgHeader);
        uint8_t *wkc_raw = pack_data(data_raw); // сначала упаковываем  все данные для вычисления их размера
        uint16_t len = wkc_raw - data_raw;      // вычисляем размер данных
        ec_moredatagrams more = (get_next() != nullptr) ? ec_moredatagrams::EC_MOREDATAGRAMS_MORE : ec_moredatagrams::EC_MOREDATAGRAMS_LAST;
        TEcatDgHeader *header_ = new(raw) TEcatDgHeader{
            command_.get_cmd(),
            idx,
            command_.get_address(),
            {.bits = {
                      .len = len,
                      .c = static_cast<uint16_t>(ec_circframe::EC_CIRCFRAME_NOT),
                      .m = static_cast<uint16_t>(more)}},
            0x0000};

        (void)header_;

        return pack_wkc(wkc_raw);
    }

    uint8_t* unpack_wkc(uint8_t *raw) {
        TEcatWkc *wkc = reinterpret_cast<TEcatWkc*>(raw);

        wkc_ = *wkc;

        return raw + sizeof(TEcatWkc);
    }

    uint8_t* unpack_data(uint8_t *raw) {
        UnpackFunctor functor{raw, command_.get_skip_unpack()};

        for_each(data_tuple_, functor);

        return unpack_wkc(functor.raw);
    }

    uint8_t* unpack_header(uint8_t *raw, uint8_t idx) {
        TEcatDgHeader *header = reinterpret_cast<TEcatDgHeader*>(raw);

        if(header->idx != idx) {
            return nullptr;
        }

        header_ = *header;

        return unpack_data(raw + sizeof(TEcatDgHeader));
    }
};

}

}

#endif /* FREE_RTOS_ETHERNET_INDUSTRY_ETH_ECAT_DATAGRAM_HPP_ */
