/** \file
 * \brief
 * Типы данных и константы, используемые для работы с EtherCAT сетью.
 */
#ifndef ETHERCAT_MASTER_ETHERCATTYPE_HPP
#define ETHERCAT_MASTER_ETHERCATTYPE_HPP

#include <cstdint>

#include <memory>
#include <array>
#include <sys/types.h>

namespace free_rtos {
    
    const std::uint16_t COUNTER_INIT = 0u;
    const std::uint16_t MIN_DATA_VECTOR_SIZE = 2u;
    const std::uint16_t ONE_BYTE_SHIFT = 8u;
    const std::uint16_t TWO_BYTE_SHIFT = 16u;
    const std::uint16_t THREE_BYTE_SHIFT = 24u;
    
    
    const std::uint16_t RESERVED_VALUE = 0;
    
    /*!
     *  Constants used to add_slaves Ethernet network when working with EtherCAT.
     */
    enum class ec_network : std::uint16_t {
        /** Port used to send EtherCAT datagrams */
        PROTOCOL_TYPE = 1u,
        ETHERCAT_PORT = 34980u
        
    };
    
    enum class sync_manager : std::uint16_t {
        SYNC_M0 = 0,
        SYNC_M1,
        SYNC_M2,
        SYNC_M3
    };
    enum class fmmu : std::uint16_t {
        FMMU0 = 0,
        FMMU1,
        FMMU2,
        FMMU3
    };
    
    enum class DataDirection : std::uint8_t {
        READ = 0x1,
        WRITE = 0x2
    };
    
    
    /**
     * Структура данных описывающая базовые настройки буффера данных, обрабатывающего меилбоксы.
     */
    struct MbxBufferProperties {
        /** Адрес в памяти подчиненного устройства для доступа на запись запросов */
        std::uint16_t mbx_write_adr;
        /** Размер буфера данных, обрабатывающего меилбоксы */
        std::uint16_t mbx_write_len;
        /** Адрес в памяти подчиненного устройства для доступа на чтение ответов */
        std::uint16_t mbx_read_adr;
        /** Размер буфера данных, обрабатывающего меилбоксы */
        std::uint16_t mbx_read_len;
        /** Адрес в памяти подчиненного устройства для доступа на чтение ответов */
        std::uint16_t pdo_output_adr;
        /** Адрес в памяти подчиненного устройства для доступа на чтение ответов */
        std::uint16_t pdo_input_adr;
    };
    
    /*!
     * EtherCAT commands types
     */
    enum ec_cmdtype : std::uint16_t {
        /** No operation */
        EC_CMD_NOP          = 0x0,
        /** Auto Increment Read */
        EC_CMD_APRD,
        /** Auto Increment Write */
        EC_CMD_APWR,
        /** Auto Increment Read Write */
        EC_CMD_APRW,
        /** Configured Address Read */
        EC_CMD_FPRD,
        /** Configured Address Write */
        EC_CMD_FPWR,
        /** Configured Address Read Write */
        EC_CMD_FPRW,
        /** Broadcast Read */
        EC_CMD_BRD,
        /** Broadcast Write */
        EC_CMD_BWR,
        /** Broadcast Read Write */
        EC_CMD_BRW,
        /** Logical Memory Read */
        EC_CMD_LRD,
        /** Logical Memory Write */
        EC_CMD_LWR,
        /** Logical Memory Read Write */
        EC_CMD_LRW,
        /** Auto Increment Read Multiple Write */
        EC_CMD_ARMW,
        /** Configured Read Multiple Write */
        EC_CMD_FRMW
        /** Reserved */
    } ;
    
    /** Ethercat registers */
    enum ec_reg_offset :  std::uint16_t
    {
        ECT_REG_TYPE        = 0x0000,
        ECT_REG_PORTDES     = 0x0007,
        ECT_REG_ESCSUP      = 0x0008,
        ECT_REG_STADR       = 0x0010,
        ECT_REG_ALIAS       = 0x0012,
        ECT_REG_DLCTL       = 0x0100,
        ECT_REG_DLPORT      = 0x0101,
        ECT_REG_DLALIAS     = 0x0103,
        ECT_REG_DLSTAT      = 0x0110,
        ECT_REG_ALCTL       = 0x0120,
        ECT_REG_ALSTAT      = 0x0130,
        ECT_REG_ALSTATCODE  = 0x0134,
        ECT_REG_PDICTL      = 0x0140,
        ECT_REG_IRQMASK     = 0x0200,
        ECT_REG_RXERR       = 0x0300,
        ECT_REG_FRXERR      = 0x0308,
        ECT_REG_EPUECNT     = 0x030C,
        ECT_REG_PECNT       = 0x030D,
        ECT_REG_PECODE      = 0x030E,
        ECT_REG_LLCNT       = 0x0310,
        ECT_REG_WD_DIV      = 0x0400,
        ECT_REG_WD_TIME_PDI = 0x0410,
        ECT_REG_WD_TIME_PD  = 0x0420,
        ECT_REG_WDCNT       = 0x0442,
        ECT_REG_EEPCFG      = 0x0500,
        ECT_REG_EEPCTL      = 0x0502,
        ECT_REG_EEPSTAT     = 0x0502,
        ECT_REG_EEPADR      = 0x0504,
        ECT_REG_EEPDAT      = 0x0508,
        ECT_REG_FMMU0       = 0x0600,
        ECT_REG_FMMU1       = ECT_REG_FMMU0 + 0x10,
        ECT_REG_FMMU2       = ECT_REG_FMMU1 + 0x10,
        ECT_REG_FMMU3       = ECT_REG_FMMU2 + 0x10,
        ECT_REG_SM0         = 0x0800,
        ECT_REG_SM1         = ECT_REG_SM0 + 0x08,
        ECT_REG_SM2         = ECT_REG_SM1 + 0x08,
        ECT_REG_SM3         = ECT_REG_SM2 + 0x08,
        ECT_REG_SM0STAT     = ECT_REG_SM0 + 0x05,
        ECT_REG_SM1STAT     = ECT_REG_SM1 + 0x05,
        ECT_REG_SM1ACT      = ECT_REG_SM1 + 0x06,
        ECT_REG_SM1CONTR    = ECT_REG_SM1 + 0x07,
        ECT_REG_DCTIME0     = 0x0900,
        ECT_REG_DCTIME1     = 0x0904,
        ECT_REG_DCTIME2     = 0x0908,
        ECT_REG_DCTIME3     = 0x090C,
        ECT_REG_DCSYSTIME   = 0x0910,
        ECT_REG_DCSOF       = 0x0918,
        ECT_REG_DCSYSOFFSET = 0x0920,
        ECT_REG_DCSYSDELAY  = 0x0928,
        ECT_REG_DCSYSDIFF   = 0x092C,
        ECT_REG_DCSPEEDCNT  = 0x0930,
        ECT_REG_DCTIMEFILT  = 0x0934,
        ECT_REG_DCCUC       = 0x0980,
        ECT_REG_DCSYNCACT   = 0x0981,
        ECT_REG_DCSTART0    = 0x0990,
        ECT_REG_DCCYCLE0    = 0x09A0,
        ECT_REG_DCCYCLE1    = 0x09A4
    };
    
    /** Ethercat registers */
    enum ec_pdo_assigment : std::uint16_t {
        rx  = 0x1C12,
        tx  = 0X1c13
    };
    
    enum MailboxesRegs {
        WRITE = 0,
        READ,
        EMPTY,
        AVAILABLE
    };
    
    /** Possible EtherCAT slave states */
    typedef enum
    {
        /** No valid state. */
        EC_STATE_NONE           = 0x00,
        /** Init state*/
        EC_STATE_INIT           = 0x01,
        /** Pre-operational. */
        EC_STATE_PRE_OP         = 0x02,
        /** Boot state*/
        EC_STATE_BOOT           = 0x03,
        /** Safe-operational. */
        EC_STATE_SAFE_OP        = 0x04,
        /** Operational */
        EC_STATE_OPERATIONAL    = 0x08,
        /** Error or ACK error */
        EC_STATE_ACK            = 0x10,
        EC_STATE_ERROR          = 0x10
    } ec_state;
    
    /** EEprom state machine read size */
    constexpr std::uint32_t EC_ESTAT_R64 = 0x0040;
    /** EEprom state machine busy flag */
    constexpr std::uint32_t EC_ESTAT_BUSY = 0x8000;
    /** EEprom state machine error flag mask */
    constexpr std::uint32_t EC_ESTAT_EMASK = 0x7800;
    /** EEprom state machine error acknowledge */
    constexpr std::uint32_t EC_ESTAT_NACK = 0x2000;
    
    /** Ethercat EEprom command types */
    typedef enum
    {
        /** No operation */
        EC_ECMD_NOP         = 0x0000,
        /** Read */
        EC_ECMD_READ        = 0x0100,
        /** Write */
        EC_ECMD_WRITE       = 0x0201,
        /** Reload */
        EC_ECMD_RELOAD      = 0x0300
    } ec_ecmdtype;
    
    /** Item offsets in SII general section */
    enum
    {
        ECT_SII_MANUF       = 0x0008,
        ECT_SII_ID          = 0x000a,
        ECT_SII_REV         = 0x000c,
        ECT_SII_BOOTRXMBX   = 0x0014,
        ECT_SII_BOOTTXMBX   = 0x0016,
        ECT_SII_MBXSIZE     = 0x0019,
        ECT_SII_TXMBXADR    = 0x001a,
        ECT_SII_RXMBXADR    = 0x0018,
        ECT_SII_MBXPROTO    = 0x001c,
        ECT_PDOOUTPUTADR    = 0x0078,
        ECT_PDOINPUTADR     = 0x007c
        
    };
    
    /*!
     * Circulating frames bit
     */
    enum class ec_circframe :  std::uint16_t {
        /** Frame is not circulating */
        EC_CIRCFRAME_NOT = 0u,
        /** Frame has circulated once */
        EC_CIRCFRAME_ONCE,
    } ;
    
    /** standard SM0 flags configuration for mailbox slaves */
    constexpr std::uint32_t EC_DEFAULTMBXSM0 = 0x00010026;
    /** standard SM1 flags configuration for mailbox slaves */
    constexpr std::uint32_t EC_DEFAULTMBXSM1 = 0x00010022;
    /** standard SM2 flags configuration for mailbox slaves */
    constexpr std::uint32_t EC_DEFAULTMBXSM2 = 0x00010074;
    /** standard SM3 flags configuration for mailbox slaves */
    constexpr std::uint32_t EC_DEFAULTMBXSM3 = 0x00010030;
    
    
    
    
    /*!
     * More EtherCAT datagrams bit
     */
    enum ec_moredatagrams :  std::uint16_t {
        /** Last EtherCAT datagram */
        EC_MOREDATAGRAMS_LAST = 0u,
        /** More EtherCAT datagrams will follow */
        EC_MOREDATAGRAMS_MORE,
    } ;
    
    
    /**  EtherCAT Frame Header :
     * <br>Length - 11 bit : Length of the EtherCAT datagrams (excl. FCS)
     * <br>Reserved - 1 bit: Reserved, 0
     * <br>Type - 4 bit : Protocol type. Only EtherCAT commands (Type = 0x1) are supported by ESCs.
     */
    struct EthercatFrameHeader {
        enum EthercatFrameHeaderSizes : std::uint16_t {
            RESERVED            = 1u,
            TYPE                = 4u,
            LENGHT              = 11u
        };
        
        
        std::uint16_t lenght : EthercatFrameHeaderSizes::LENGHT;
        std::uint16_t res : EthercatFrameHeaderSizes::RESERVED;
        std::uint16_t type : EthercatFrameHeaderSizes::TYPE;
    };
    
    constexpr static std::uint16_t write_synch_manager_buffer_size = 256;
    constexpr static std::uint16_t read_synch_manager_buffer_size = 256;
    
} // namespace ethercat


#endif //ETHERCAT_MASTER_ETHERCATTYPE_HPP
